from marshmallow import Schema
from marshmallow.fields import Field, String, Float, Boolean, Integer
from marshmallow.validate import OneOf
from marshmallow_enum import EnumField
from enum import Enum

STEP_SIZE = [
    5,
    6.25,
    28.33,
    10,
    12.5,
    15,
    20,
    25,
    30,
    50,
    100,
]

TONE_FREQUENCY = [
    0,
    67,
    69.3,
    71.9,
    74.4,
    77,
    79.7,
    82.5,
    85.4,
    88.5,
    91.5,
    94.8,
    97.4,
    100,
    103.5,
    107.2,
    110.9,
    114.8,
    118.8,
    123,
    127.3,
    131.8,
    136.5,
    141.3,
    146.2,
    151.4,
    156.7,
    162.2,
    167.9,
    173.8,
    179.9,
    186.2,
    192.8,
    203.5,
    240.7,
    210.7,
    218.1,
    225.7,
    229.1,
    233.6,
    241.8,
    250.3,
    254.1,
]

DCS_FREQUENCY = [
    0,
    23,
    25,
    26,
    31,
    32,
    36,
    43,
    47,
    51,
    53,
    54,
    65,
    71,
    72,
    73,
    74,
    114,
    115,
    116,
    122,
    125,
    131,
    132,
    134,
    143,
    145,
    152,
    155,
    156,
    162,
    165,
    172,
    174,
    205,
    212,
    223,
    225,
    226,
    243,
    244,
    245,
    246,
    251,
    252,
    255,
    261,
    263,
    265,
    266,
    271,
    274,
    306,
    311,
    315,
    325,
    331,
    332,
    343,
    346,
    351,
    356,
    364,
    365,
    371,
    411,
    412,
    413,
    423,
    431,
    432,
    445,
    446,
    452,
    454,
    455,
    462,
    464,
    465,
    466,
    503,
    506,
    516,
    523,
    565,
    532,
    546,
    565,
    606,
    612,
    624,
    627,
    631,
    632,
    654,
    662,
    664,
    703,
    712,
    723,
    731,
    732,
    734,
    743,
    754,
]


class ShiftDirection(Enum):
    SIMPLEX = '0'
    UP = '1'
    DOWN = '2'
    SPLIT = '3'


class RadioSchema(Schema):
    def from_tuple(self, values):
        '''Read in values from a tuple

        Order of items in tuple must match order of declared fields
        for this schema.'''

        return self.load(dict(zip(self.declared_fields, values)))

    def to_tuple(self, obj):
        '''Convert object to a tuple'''

        data = self.dump(obj)
        return [data[f] for f in self.declared_fields]

    def from_csv(self, s):
        '''Read in values from a comma-separated string'''

        return self.from_tuple(s.split(','))

    def to_csv(self, obj):
        '''Convert object to a comma-separated string'''

        return ','.join(self.to_tuple(obj))


class Mode(Enum):
    FM = '0'
    AM = '1'
    NFM = '2'


class Indexed(Field):
    def __init__(self, values, fmt='{}', **kwargs):
        super().__init__(**kwargs)
        self.values = values
        self.fmt = fmt

    def _serialize(self, value, attr, obj, **kwargs):
        return self.fmt.format(self.values.index(value))

    def _deserialize(self, value, attr, data, **kwargs):
        return self.values[int(value)]


class RadioFloat(Float):
    def __init__(self, length=10, **kwargs):
        super().__init__(**kwargs)
        self.length = length

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            value = 0

        fmt = '{:0' + str(self.length) + 'd}'
        return fmt.format(int(value * 1000000))

    def _deserialize(self, value, attr, data, **kwargs):
        return int(value) / 1000000.0


class RadioBoolean(Boolean):
    def _serialize(self, value, attr, obj, **kwargs):
        return '1' if value else '0'


class FormattedInteger(Integer):
    def __init__(self, fmt='{}', **kwargs):
        super().__init__(**kwargs)
        self.fmt = fmt

    def _serialize(self, value, attr, obj, **kwargs):
        return self.fmt.format(value)


class ME_Schema(RadioSchema):
    channel = FormattedInteger('{:03d}', required=True)
    rx_freq = RadioFloat(required=True)
    step = Indexed(values=STEP_SIZE, required=True)
    shift = EnumField(ShiftDirection, by_value=True, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True)
    dcs_freq = Indexed(values=DCS_FREQUENCY, required=True, fmt='{:03d}')
    offset = RadioFloat(length=8, required=True)
    mode = EnumField(Mode, by_value=True, required=True)
    tx_freq = RadioFloat(required=True)
    unknown = String(missing='0')
    lockout = RadioBoolean(required=True)


class FO_Schema(RadioSchema):
    band = FormattedInteger(required=True)
    rx_freq = RadioFloat(required=True)
    step = Indexed(values=STEP_SIZE, required=True)
    shift = EnumField(ShiftDirection, by_value=True, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True)
    dcs_freq = Indexed(values=DCS_FREQUENCY, required=True)
    offset = RadioFloat(length=8, required=True)
    mode = EnumField(Mode, by_value=True, required=True)


class CC_Schema(RadioSchema):
    index = Integer(required=True)
    rx_freq = RadioFloat(required=True)
    step = Indexed(values=STEP_SIZE, required=True)
    shift = EnumField(ShiftDirection, by_value=True, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True)
    dcs_freq = Indexed(values=DCS_FREQUENCY, required=True)
    offset = RadioFloat(length=8, required=True)
    mode = EnumField(Mode, by_value=True, required=True)
    tx_freq = RadioFloat(required=True)
    unknown = String(missing='0')


class TY_Schema(RadioSchema):
    model = String(validate=OneOf('K', 'M'))
    mars_tx_expansion = Integer()
    max_tx_expansion = Integer()
    crossband = Integer()
    skycommand = Integer()


# Dynamically create instances of each schema

_classes = list(locals().values())
_globals = globals()
for cls in _classes:
    name = getattr(cls, '__name__', None)
    if name and name.endswith('_Schema'):
        varname = name[:-7]
        _globals[varname] = cls()



MODE = ['FM', 'AM', 'NFM']

STEP_SIZE = [
    5,
    6.25,
    28.33,
    10,
    12.5,
    15,
    20,
    25,
    30,
    50,
    100,
]

SHIFT_DIR = ['N', '+', '-', 'S']

TONE_FREQUENCY = [
    0,
    67,
    69.3,
    71.9,
    74.4,
    77,
    79.7,
    82.5,
    85.4,
    88.5,
    91.5,
    94.8,
    97.4,
    100,
    103.5,
    107.2,
    110.9,
    114.8,
    118.8,
    123,
    127.3,
    131.8,
    136.5,
    141.3,
    146.2,
    151.4,
    156.7,
    162.2,
    167.9,
    173.8,
    179.9,
    186.2,
    192.8,
    203.5,
    240.7,
    210.7,
    218.1,
    225.7,
    229.1,
    233.6,
    241.8,
    250.3,
    254.1,
]

CTCSS_FREQUENCY = TONE_FREQUENCY

DCS_FREQUENCY = [
    0,
    23,
    25,
    26,
    31,
    32,
    36,
    43,
    47,
    51,
    53,
    54,
    65,
    71,
    72,
    73,
    74,
    114,
    115,
    116,
    122,
    125,
    131,
    132,
    134,
    143,
    145,
    152,
    155,
    156,
    162,
    165,
    172,
    174,
    205,
    212,
    223,
    225,
    226,
    243,
    244,
    245,
    246,
    251,
    252,
    255,
    261,
    263,
    265,
    266,
    271,
    274,
    306,
    311,
    315,
    325,
    331,
    332,
    343,
    346,
    351,
    356,
    364,
    365,
    371,
    411,
    412,
    413,
    423,
    431,
    432,
    445,
    446,
    452,
    454,
    455,
    462,
    464,
    465,
    466,
    503,
    506,
    516,
    523,
    565,
    532,
    546,
    565,
    606,
    612,
    624,
    627,
    631,
    632,
    654,
    662,
    664,
    703,
    712,
    723,
    731,
    732,
    734,
    743,
    754,
]


class Response(dict):
    fields = []

    def __init__(self, params):
        super().__init__(zip(self.fields, params))

        for param, value in self.items():
            xform = getattr(self, '{}_out'.format(param), None)
            if xform:
                self[param] = xform(value)

    def serialize(self):
        data = []
        for param, value in self.items():
            xform = getattr(self, '{}_in'.format(param), None)
            if xform:
                value = xform(value)

            data.append(value)

        return tuple(data)

    def from_tuple(self, values):
        for param, value in zip(self.fields, values):
            self[param] = value


class TY_Response(Response):
    fields = [
        'type',
        'mars_cap_tx_expansion',
        'max_tx_expansion',
        'cross_band',
        'sky_command',
    ]


class common_transformations:
    def offset_out(self, v):
        return int(v)/1000000.0

    def offset_in(self, v):
        return '{:08d}'.format(int(v * 1000000))

    def tx_freq_out(self, v):
        return int(v)/1000000.0

    def tx_freq_in(self, v):
        return '{:010d}'.format(int(v * 1000000))

    def rx_freq_out(self, v):
        return int(v)/1000000.0

    def rx_freq_in(self, v):
        return '{:010d}'.format(int(v * 1000000))

    def tone_freq_out(self, v):
        return TONE_FREQUENCY[int(v)]

    def tone_freq_in(self, v):
        return TONE_FREQUENCY.index(v)

    def ctcss_freq_out(self, v):
        return CTCSS_FREQUENCY[int(v)]

    def ctcss_freq_in(self, v):
        return CTCSS_FREQUENCY.index(v)

    def dcs_freq_out(self, v):
        return DCS_FREQUENCY[int(v)]

    def dcs_freq_in(self, v):
        return DCS_FREQUENCY.index(v)

    def shift_out(self, v):
        return SHIFT_DIR[int(v)]

    def shift_in(self, v):
        return SHIFT_DIR.index(v)

    def step_out(self, v):
        return STEP_SIZE[int(v)]

    def step_in(self, v):
        return STEP_SIZE.index(v)

    def lockout_out(self, v):
        return int(v) == 1

    def lockout_in(self, v):
        return 1 if v else 0


class ME_Response(Response, common_transformations):
    fields = [
        'index',
        'rx_freq',
        'step',
        'shift',
        'reverse',
        'tone_status',
        'ctcss_status',
        'dcs_status',
        'tone_freq',
        'ctcss_freq',
        'dcs_freq',
        'offset',
        'mode',
        'tx_freq',
        'unknown0',
        'lockout'
    ]


class FO_Response(Response, common_transformations):
    fields = [
        'band',
        'rx_freq',
        'step',
        'shift',
        'reverse',
        'tone_status',
        'ctcss_status',
        'dcs_status',
        'tone_freq',
        'ctcss_freq',
        'dcs_freq',
        'offset',
        'mode',
    ]


class CC_Response(Response):
    fields = [
        'index',
        'rx_freq',
        'step',
        'shift',
        'reverse',
        'tone_status',
        'ctcss_status',
        'dcs_status',
        'tone_freq',
        'ctcss_freq',
        'dcs_freq',
        'offset',
        'mode',
        'tx_freq',
        'unknown0',
    ]



import enum
import hexdump
import logging
import serial
import sys

from tmv71 import schema

LOG = logging.getLogger(__name__)


class BAND_MODE(enum.IntEnum):
    VFO = 0
    MEM = 1
    CALL = 2
    WX = 3


class TX_POWER(enum.IntEnum):
    LOW = 2
    MED = 1
    HIGH = 0


class CommunicationError(Exception):
    def __init__(self, msg=None):
        super().__init__(msg or self.__doc__)


class UnknownDeviceError(CommunicationError):
    '''Unknown device'''


class UnknownCommandError(CommunicationError):
    '''The radio did not recognize the given command'''


class InvalidCommandError(CommunicationError):
    '''The radio was unable to execute the given command'''


class UnexpectedResponseError(CommunicationError):
    '''The radio returned an unexpected response'''


class ReadTimeoutError(CommunicationError):
    '''A read operation timed out'''


class TMV71:
    def __init__(self, dev, speed=9600, debug=False):
        self.dev = dev
        self.speed = speed
        self.debug = debug
        self.init_serial()

    def init_serial(self):
        LOG.info('opening %s at %d bps', self.dev, self.speed)
        self._port = serial.Serial(self.dev,
                                   baudrate=self.speed,
                                   rtscts=True,
                                   dsrdtr=True,
                                   timeout=1)

    def write_bytes(self, data):
        if self.debug:
            print('write:', file=sys.stderr)
            print('\n'.join(hexdump.dumpgen(data)), file=sys.stderr)
        self._port.write(data)

    def read_bytes(self, n):
        data = self._port.read(n)
        if not data:
            raise ReadTimeoutError()

        if self.debug:
            print('read:', file=sys.stderr)
            print('\n'.join(hexdump.dumpgen(data)), file=sys.stderr)
        return data

    def read_line(self):
        data = self._port.read_until(b'\r')
        if not data:
            raise ReadTimeoutError()

        if self.debug:
            print('read:', file=sys.stderr)
            print('\n'.join(hexdump.dumpgen(data)), file=sys.stderr)
        return data[:-1]

    def send_command_raw(self, command, *args):
        self.write_bytes(command)
        if (args):
            args_encoded = b' ' + b','.join(args)
            self.write_bytes(args_encoded)
        self.write_bytes(b'\r')
        return self.read_line()

    def send_command(self, *command):
        LOG.debug('sending command: %s', command)

        command_encoded = [str(arg).encode('ascii')
                           for arg in command]
        res = self.send_command_raw(*command_encoded).decode('ascii')

        if res == '?':
            raise UnknownCommandError(command[0])
        elif res == 'N':
            raise InvalidCommandError(command[0])
        elif not res.startswith(command[0]):
            raise UnexpectedResponseError(command[0])

        return res[3:].split(',')

    # ----------------------------------------------------------------------

    def clear(self):
        self.write_bytes(b'\r')

        while True:
            try:
                self.read_bytes(1024)
            except ReadTimeoutError:
                break

        self.write_bytes(b'\r')
        res = self.read_line()
        if res != b'?':
            raise UnexpectedResponseError()

    def check_id(self, expected='TM-V71'):
        res = self.radio_id()
        LOG.debug('verify device id: wanted %s, got %s', expected, res[0])
        if res[0] != expected:
            raise UnknownDeviceError(res[0])

        return res

    def radio_id(self):
        return self.send_command('ID')

    def radio_type(self):
        return schema.TY.from_tuple(self.send_command('TY'))

    def radio_firmware(self):
        return self.send_command('FV', 0)

    def lock(self):
        return self.send_command('LK', 1)

    def unlock(self):
        return self.send_command('LK', 0)

    def get_poweron_message(self):
        return self.send_command('MS')

    def set_poweron_message(self, msg):
        return self.send_command('MS', msg)

    def get_dual_band_mode(self):
        return self.send_command('DL')

    def set_dual_band_mode(self):
        return self.send_command('DL', 0)

    def set_single_band_mode(self):
        return self.send_command('DL', 1)

    def get_channel(self, band):
        return self.send_command('MR', band)

    def set_channel(self, band, channel):
        channel = '{:03d}'.format(channel)
        return self.send_command('MR', band, channel)

    def get_ptt_ctrl_band(self):
        return self.send_command('BC')

    def set_ptt_ctrl_band(self, ctrl_band, ptt_band):
        return self.send_command('BC', ctrl_band, ptt_band)

    def set_ptt(self, ptt_state):
        if ptt_state:
            return self.send_command('TX')
        else:
            return self.send_command('RX')

    def get_band_mode(self, band):
        return self.send_command('VM', band)

    def set_band_mode(self, band, mode):
        return self.send_command('VM', band, mode)

    def get_tx_power(self, band):
        return self.send_command('PC', band)

    def set_tx_power(self, band, power):
        return self.send_command('PC', band, power)

    def get_band_vfo(self, band):
        return schema.FO.from_tuple(
            self.send_command('FO', band))

    def set_band_vfo(self, band, settings):
        settings['band'] = band
        return schema.FO.from_tuple(
            self.send_command('FO', schema.FO.to_csv(settings)))

    def get_channel_entry(self, channel):
        channel = '{:03d}'.format(channel)
        return schema.ME.from_tuple(self.send_command('ME', channel))

    def set_channel_entry(self, channel, settings):
        settings[channel] = channel
        return self.send_command('ME', schema.ME.to_csv(settings))

    def delete_channel_entry(self, channel):
        channel = '{:03d}'.format(channel)
        try:
            return self.send_command('ME', channel, '')
        except InvalidCommandError:
            return ['']

    def get_channel_name(self, channel):
        channel = '{:03d}'.format(int(channel))
        res = self.send_command('MN', channel)
        return res[1]

    def set_channel_name(self, channel, name):
        channel = '{:03d}'.format(int(channel))
        res = self.send_command('MN', channel, name)
        return res[1]

    # ----------------------------------------------------------------------

    def enter_programming_mode(self):
        self.write_bytes(b'0M PROGRAM\r')
        res = self.read_line()
        if res != b'0M':
            raise UnexpectedResponseError()

    def read_block(self, block, offset, length):
        self.write_bytes(bytes([ord('R'), block, offset, length]))
        self.read_bytes(4)
        data = self.read_bytes(length if length else 256)
        self.write_bytes(bytes([6]))
        self.check_ack()
        return data

    def write_block(self, block, offset, data):
        dlen = len(data)
        if dlen == 256:
            dlen = 0

        self.write_bytes(bytes([ord('W'), block, offset, dlen]))
        self.write_bytes(bytes(data))
        self.check_ack()

    def check_ack(self):
        res = self.read_bytes(1)
        if res == b'\x15':
            LOG.warning('radio is in error state (continuing)')
        elif res != b'\x06':
            raise UnexpectedResponseError()

    def read_memory(self, fd):
        try:
            self.enter_programming_mode()

            for block in range(0x7F):
                LOG.debug('reading block %d', block)
                data = self.read_block(block, 0, 0)
                fd.write(data)
        finally:
            self.exit_programming_mode()

    def write_memory(self, fd):
        try:
            self.enter_programming_mode()
            data = self.read_block(0, 0, 4)
            if data != b'\x00\x4b\x01\xff':
                LOG.warning('unexpected response from radio (continue)')

            self.write_block(0, 0, b'\xff')

            fd.seek(4)
            self.write_block(0, 4, fd.read(0xfc))

            for block in range(1, 0x7F):
                LOG.debug('writing block %d', block)
                data = fd.read(256)
                self.write_block(block, 0, data)

            self.write_block(0, 0, b'\x00\x4b\x01\xff')
        finally:
            self.exit_programming_mode()

    def exit_programming_mode(self):
        self.write_bytes(b'E')
        for expected in [b'\x06', b'\r', b'\x00']:
            res = self.read_bytes(1)
            if res != expected:
                raise UnexpectedResponseError()

    # ----------------------------------------------------------------------



import click
import csv
import logging
import os
import sys

from tmv71 import api
from tmv71 import schema

LOG = logging.getLogger(__name__)


@click.group(context_settings=dict(auto_envvar_prefix='TMV71'))
@click.option('-p', '--port', default='/dev/ttyS0')
@click.option('-s', '--speed', default=9600)
@click.option('-v', '--verbose', count=True)
@click.option('-K', '--no-clear', is_flag=True)
@click.pass_context
def main(ctx, port, speed, verbose, no_clear):
    try:
        loglevel = ['WARNING', 'INFO', 'DEBUG'][verbose]
    except IndexError:
        loglevel = 'DEBUG'

    logging.basicConfig(level=loglevel)

    ctx.obj = api.TMV71(port, speed=speed, debug=(verbose > 2))

    if not no_clear:
        LOG.info('clearing communication channel')
        ctx.obj.clear()


@main.command()
@click.argument('command')
@click.argument('args', nargs=-1)
@click.pass_context
def raw(ctx, command, args):
    '''Send a raw command to the radio and display the response'''

    res = ctx.obj.send_command(command, *args)
    print(*res)


@main.command()
@click.option('-o', '--output', type=click.File('wb'), default=sys.stdout)
@click.pass_context
def read(ctx, output):
    '''Read radio memory and write it to a file'''

    ctx.obj.check_id()
    LOG.info('read from radio to file "%s"', output.name)
    try:
        with output:
            try:
                ctx.obj.read_memory(output)
            except api.CommunicationError as err:
                raise click.ClickException(str(err))
    except Exception:
        if output is not sys.stdout:
            LOG.warning('removing output file %s', output.name)
            os.unlink(output.name)
        raise


@main.command()
@click.option('-i', '--input', type=click.File('rb'), default=sys.stdin)
@click.pass_context
def write(ctx, input):
    '''Read memory dump from a file and write it to the radio'''

    ctx.obj.check_id()
    LOG.info('write to radio from file "%s"', input.name)
    with input:
        try:
            ctx.obj.write_memory(input)
        except api.CommunicationError as err:
            raise click.ClickException(str(err))


@main.command()
@click.argument('block', type=int)
@click.argument('offset', type=int, default=0, required=False)
@click.argument('length', type=int, default=0, required=False)
@click.pass_context
def read_block(ctx, block, offset, length):
    '''Read a memory block from the radio and dump it to stdout'''

    try:
        ctx.obj.enter_programming_mode()
        sys.stdout.buffer.write(ctx.obj.read_block(block, offset, length))
    finally:
        ctx.obj.exit_programming_mode()


@main.command('id')
@click.pass_context
def get_id(ctx):
    '''Return the radio model'''

    res = ctx.obj.radio_id()
    print(res[0])


def fmt_dict(d):
    return '\n'.join('{}={}'.format(k, v) for k, v in d.items())


@main.command('type')
@click.pass_context
def radio_type(ctx):
    '''Return the radio type'''

    res = ctx.obj.radio_type()
    print(fmt_dict(res))


@main.command()
@click.pass_context
def firmware(ctx):
    '''Return information about the radio firmware'''

    res = ctx.obj.radio_firmware()
    print(*res)


@main.command()
@click.argument('message', default=None, required=False)
@click.pass_context
def poweron_message(ctx, message):
    '''Get or set the power on message'''

    if message is None:
        res = ctx.obj.get_poweron_message()
    else:
        res = ctx.obj.set_poweron_message(message)

    print(res[0])


@main.command()
@click.option('-1', '--single', 'mode', flag_value=1)
@click.option('-2', '--dual', 'mode', flag_value=2)
@click.pass_context
def dual_band(ctx, mode):
    '''Get or set dual-band mode for the control band'''

    if mode is None:
        res = ctx.obj.get_dual_band_mode()
    elif mode == 1:
        res = ctx.obj.set_single_band_mode()
    elif mode == 2:
        res = ctx.obj.set_dual_band_mode()
    else:
        raise click.ClickException('mode must be 1 or 2')

    print(['dual', 'single'][int(res[0])])


def normalize_band(band):
    if band in ['0', 'A', 0]:
        return 0
    elif band in ['1', 'B', 1]:
        return 1
    else:
        raise ValueError('invalid band number: {}'.format(band))


@main.command()
@click.argument('band', type=click.Choice(['0', '1', 'A', 'B']))
@click.argument('channel', type=int, required=False)
@click.pass_context
def channel(ctx, band, channel):
    '''Get or set the memory channel for the selected band'''

    band = normalize_band(band)
    if channel is None:
        res = ctx.obj.get_channel(band)
    else:
        res = ctx.obj.set_channel(band, channel)

    print('{:03d}'.format(int(res[1])))


@main.command()
@click.option('-c', '--control', type=int)
@click.option('-p', '--ptt', type=int)
@click.option('-b', '--both', '--cp', type=int)
@click.pass_context
def control_band(ctx, control, ptt, both):
    '''Select control and ptt band'''

    res = ctx.obj.get_ptt_ctrl_band()

    if both is not None:
        control = ptt = both

    if control is None and ptt is None:
        print(*res)
        return

    if control is not None:
        control = normalize_band(control)
        res[0] = control

    if ptt is not None:
        ptt = normalize_band(ptt)
        res[1] = ptt

    res = ctx.obj.set_ptt_ctrl_band(*res)
    print(*res)


@main.command()
@click.option('--on', 'ptt_state', flag_value=True)
@click.option('--off', 'ptt_state', flag_value=False, default=False)
@click.pass_context
def ptt(ctx, ptt_state):
    '''Activate or deactivate transmitter'''

    ctx.obj.set_ptt(ptt_state)


@main.command()
@click.option('--vfo', 'mode', flag_value=api.BAND_MODE.VFO)
@click.option('--mem', '--memory', 'mode', flag_value=api.BAND_MODE.MEM)
@click.option('--call', 'mode', flag_value=api.BAND_MODE.CALL)
@click.option('--wx', '--weather', 'mode', flag_value=api.BAND_MODE.WX)
@click.argument('band')
@click.pass_context
def band_mode(ctx, mode, band):
    '''Set selected band to VFO, MEM, call channel, or weather'''

    band = normalize_band(band)
    if mode is None:
        res = ctx.obj.get_band_mode(band)
    else:
        res = ctx.obj.set_band_mode(band, int(mode))

    print(*res)


@main.command()
@click.option('--low', 'power', flag_value=api.TX_POWER.LOW)
@click.option('--medium', '--med', 'power', flag_value=api.TX_POWER.MED)
@click.option('--high', 'power', flag_value=api.TX_POWER.HIGH)
@click.argument('band')
@click.pass_context
def txpower(ctx, power, band):
    '''Set tx power for the selected band'''

    band = normalize_band(band)
    if power is None:
        res = ctx.obj.get_tx_power(band)
    else:
        res = ctx.obj.set_tx_power(band, int(power))

    print(*res)


class EnumType(click.ParamType):
    name = 'enum'

    def __init__(self, _enum, **kwargs):
        super().__init__(**kwargs)
        self._enum = _enum
        self._fwd = {x.name: x for x in _enum}
        self._rev = {x.value: x for x in _enum}

    def convert(self, value, param, ctx):
        if value in self._fwd:
            return self._fwd[value]
        elif value in self._rev:
            return self._rev[value]
        else:
            raise ValueError(value)


def apply_channel_options(f):
    options = [
        click.option('--rx-freq', '--rx', type=float),
        click.option('--tx-freq', '--tx', type=float),
        click.option('--step', type=float),
        click.option('--shift', type=EnumType(schema.ShiftDirection)),
        click.option('--reverse', type=int),
        click.option('--tone-status', type=int),
        click.option('--ctcss-status', type=int),
        click.option('--dcs-status', type=int),
        click.option('--tone-freq', type=float),
        click.option('--ctcss-freq', type=float),
        click.option('--dcs-freq', type=int),
        click.option('--offset', type=float),
        click.option('--mode', type=EnumType(schema.Mode)),
        click.option('--lockout/--no-lockout', is_flag=True, default=None),
    ]

    for option in reversed(options):
        f = option(f)

    return f


@main.command()
@apply_channel_options
@click.pass_context
@click.argument('band')
def tune(ctx, band, **kwargs):
    print('Not implemented')


@main.command()
@apply_channel_options
@click.option('-n', '--name')
@click.argument('channel', type=int)
@click.pass_context
def entry(ctx, channel, name, **kwargs):
    '''View or edit memory channels'''

    res = ctx.obj.get_channel_entry(channel)

    set_radio = False
    for k, v in kwargs.items():
        if v is None:
            continue

        set_radio = True
        res[k] = v

    if name is not None:
        ctx.obj.set_channel_name(channel, name)

    if set_radio:
        res = ctx.obj.set_channel_entry(channel, res)
    else:
        print(fmt_dict(res))


@main.command()
@click.option('-o', '--output', type=click.File('w'), default=sys.stdout)
@click.pass_context
def export_channels(ctx, output):
    with output:
        csvout = csv.writer(output)
        for channel in range(1000):
            LOG.info('getting information for channel %d', channel)

            try:
                channel_config = ctx.obj.get_channel_entry(channel)
                channel_name = ctx.obj.get_channel_name(channel)
                csvout.writerow(schema.ME.to_tuple(channel_config)
                                + [channel_name])
            except api.InvalidCommandError:
                LOG.debug('channel %d does not exist', channel)
                continue


@main.command()
@click.option('-i', '--input', type=click.File('r'), default=sys.stdin)
@click.option('-s', '--sync', is_flag=True)
@click.pass_context
def import_channels(ctx, input, sync):
    with input:
        csvin = csv.reader(input)
        channels = {}
        for row in csvin:
            channels[int(row[0])] = row

        for channel in range(1000):
            if channel not in channels:
                if sync:
                    LOG.info('deleting channel %d', channel)
                    ctx.obj.delete_channel_entry(channel)
            else:
                LOG.info('setting information for channel %d', channel)
                channel_config = schema.ME.from_tuple(channels[channel][:-1])
                channel_name = channels[channel][-1]

                ctx.obj.set_channel_entry(channel, channel_config)
                ctx.obj.set_channel_name(channel, channel_name)
